/*
 * Simple packet capturing and network monitoring tool.
 * 
 * progfuncs.c - main program functions.
 *
 * Written by Alexander Hauser
 * lex@lufix.net
 * http://www.lufix.net/
 * http://www.sac.cc/
 *
 * */

#include "nsniff.h"

/* opens a pcap device and compiles in the bpf-code if specified */
void open_pcap() {
	char errbuf[PCAP_ERRBUF_SIZE], str_localnet[INET_ADDRSTRLEN], str_netmask[INET_ADDRSTRLEN];
	uint32_t localnet, netmask;
	struct bpf_program fcode;
		
	if (device == NULL)
		if ((device = pcap_lookupdev(errbuf)) == NULL) {
			printf("pcap_lookupdev() error: %s\n", errbuf);
			exit(1);
		}
	printf("listening on device %s\n", device);
	if ((pd = pcap_open_live(device, snaplen, 1, 0, errbuf)) == NULL) {
		printf("pcap_open_live() error: %s\n" , errbuf);
		exit(1);
	}
	if ((datalink = pcap_datalink(pd)) < 0) {
		printf("pcap_datalink() error: %s\n", pcap_geterr(pd));
		exit(1);
	}
	if (pcap_lookupnet(device, &localnet, &netmask, errbuf) < 0) {
		printf("pcap_lookupnet() error: %s\n", errbuf);
		exit(1);
	}
	if (compile) {
		if (pcap_compile(pd, &fcode, compstr, 0, netmask) < 0) {
			printf("pcap_compile() error: %s\n", errbuf);
			exit(1);
		}
		if (pcap_setfilter(pd, &fcode) < 0) {
			printf("pcap_setfilter() error: %s\n", errbuf);
			exit(1);
		}
	}
	printf("well, to exit press CTRL-C\n\n");
	switch (datalink) {
		case DLT_NULL:
			if (verbose)
				printf("datalink is loopback (4 bytes frame header)\n");
			enet_hdr = 4;
			break;
		case DLT_EN10MB:
			if (verbose)
				printf("datalink is 10/100 ethernet (14 bytes frame header)\n");
			enet_hdr = 14;
			break;
								
		case DLT_SLIP:
			if (verbose)
				printf("datalink is SLIP (24 bytes frame header)\n");
			enet_hdr = 24;
			break;

		case DLT_PPP:
			if (verbose)
				printf("datalink is PPP (24 bytes frame header)\n");
			enet_hdr = 24;
			break;
			
		default:
			printf("error: unknown ethernet type\n");
			exit(1);
	}
	if (procraw) {
		if ((pdt = pcap_dump_open(pd, "-")) != 0)
			printf("going to print the packets as RAW data too\n");
		else {
			printf("pcap_dump_open() error: %s\n", errbuf);
			exit(1);
		}
	}
	if (verbose) {
		printf("snaplen is set to %d\n", snaplen);
		inet_ntop(AF_INET, &localnet, str_localnet, sizeof(str_localnet));
		inet_ntop(AF_INET, &netmask, str_netmask, sizeof(str_netmask));
		printf("localnet is %s\n", str_localnet);
		printf("netmask is %s\n", str_netmask);
		if (compile)
			printf("compiled bpf-code is '%s'\n", compstr);
		if (counto)
			printf("going to catch %d packets and exit\n", counto);
		printf("\n");
	}
	if (!verbose && procraw) printf("\n");
		
	return;
}

/* main read loop */
void readloop() {
	int len, i;
	char *ptr;
	struct pcap_pkthdr hdr;
	struct ether_header *ether;
		
	for (;;) {
		while ((ptr = (char *) pcap_next(pd, &hdr)) == NULL);
		len = hdr.caplen;
		 
		if (counto)
			count++;
		 
		if (datalink = DLT_EN10MB) {
			ether = (struct ether_header *) ptr;
			print_time(&hdr.ts);
			 
			if (procraw) {
				printf("\n--------------------------------------------------\n");
				pcap_dump((u_char *) pdt, &hdr, ptr);
				printf("\n--------------------------------------------------\n");
			} else
				printf("\n");
			 
			switch (ntohs(ether->ether_type)) {
				case ETHERTYPE_IP:
					if (etherp) {
						printf("%d-byte frame catched. ", len);
						print_etheradrs(ether->ether_shost, ether->ether_dhost);
					}
					ip_process(ptr + enet_hdr, len - enet_hdr);
					break;
					
				case ETHERTYPE_IP6:
					if (etherp) {
						printf("%d-byte frame catched. ", len);
						print_etheradrs(ether->ether_shost, ether->ether_dhost);
					}
					ip6_process(ptr + enet_hdr, len - enet_hdr);	
					break;
					
				case ETHERTYPE_ARP:
					if (etherp) {
						printf("%d-byte frame catched. ", len);
						print_etheradrs(ether->ether_shost, ether->ether_dhost);
					}
					arp_process(ptr + enet_hdr, ether->ether_shost, ether->ether_dhost);
					break;
					
				case ETHERTYPE_REVARP:
					if (etherp) {
						printf("%d-byte frame catched. ", len);
						print_etheradrs(ether->ether_shost, ether->ether_dhost);
					}
					break;
					
				default:
					scanEtherProt(ntohs(ether->ether_type));
					break;
			}
		}
		checkcount();
		printf("\n -\n");
		fflush(stdout);
	}
	 
	return;
}

/* checks if the port belongs to a known application-layer-protocol */
void checkport(uint16_t source, uint16_t dest, char *ptr) {
	if (source==PORT_DNS || dest==PORT_DNS)
		dns_process(ptr);
	 
	return;
}

/* prints MAC addresses */
void print_etheradrs(uint8_t shost[6], uint8_t dhost[6]) {
	int i;
	 
	for (i=0; i<6; i++)
		if (i == 5)
			printf("%x", shost[i]);
		else
			printf("%x:", shost[i]);
	 
	if (dhost != 0) {
		printf(" >> ");
		for (i=0; i<6; i++)
			if (i == 5)
				printf("%x", dhost[i]);
			else
				printf("%x:", dhost[i]);
	}
	printf("\n");
	fflush(stdout);
	 
	return;
}

/* prints IPv6 address */
void print_ip6addr(const char *ap) {
	struct in6_addr addr;
	char buf[INET6_ADDRSTRLEN];
	 
	memcpy(&addr, ap, sizeof(addr));
	inet_ntop(AF_INET6, &addr, buf, sizeof(buf));
	printf("%s", buf);
		
	return;
}

/* extracts the bpf-code out of argv */
char *copy_argv(char **argv) {
	char **p;
	int len = 0;
	char *buf, *src, *dst;
	 
	p = argv;
	if (*p == 0)
		return 0;
	while (*p)
		len += strlen(*p++) + 1;
	buf = (char *)malloc(len);
	p = argv;
	dst = buf;
	while ((src = *p++) != NULL) {
		while ((*dst++ = *src++) != '\0');
		dst[-1] = ' ';
	}
	dst[-1] = '\0';
	 
	return buf;
}

/* prints precise time */
void print_time(struct timeval *tvp) {
	time_t t;
	register int daytime, dif, s;
	register struct tm *gmt, *local;
	struct tm bgmt;
	 
	t = time(NULL);
	gmt = &bgmt;
	*gmt = *gmtime(&t);
	local = localtime(&t);
	daytime = (local->tm_hour - gmt->tm_hour) * 60 * 60 +
		(local->tm_min - gmt->tm_min) * 60;
	dif = local->tm_year - gmt->tm_year;
	if (dif == 0)
		dif = local->tm_yday - gmt->tm_yday;
	daytime += dif * 24 * 60 * 60;
	 
	s = (tvp->tv_sec + daytime) % 86400;
	printf("<%02d:%02d:%02d.%06u> ", s/3600, (s%3600)/60, s%60, tvp->tv_usec);
	 
	return;
}

/* checks if specified count of captured packets has been reached */
void checkcount() {
	if (counto)
		if (counto == count) {
			printf("\n\nWell, catched %d packets and saying good-bye\n", count);
			cleanup();
		}
	return;
}

/* prints a goodbye banner and exits */
void cleanup() {
	struct pcap_stat stats;

	fflush(stdout);
	if (verbose) {
		if (pcap_stats(pd, &stats) < 0) {
			printf("pcap_stat() error: %s\n", pcap_geterr(pd));
			exit(1);
		}
		printf("\n%d packets received by filter\n", stats.ps_recv);
			printf("%d packets dropped by kernel\n", stats.ps_drop);
	}
		
	printf("\nhope it helped you ;-)\nwritten by Alexander Hauser\nlex@lufix.net\nhttp://www.lufix.net/\nhttp://www.sac.cc/\n\n");
	exit(0);
}

/* prints usage-information */
void usage(const char *msg) {
	if (msg[0] != 0) {
		printf("%s\n", msg);
		exit(1);
	}
	 
	printf("usage: ./nsniff [options] [bpf-program]\n");
	printf("options:  -c              exit after reception of specified count of packets\n");
	printf("          -e              let nSniff print link-layer headers\n");
	printf("          -h              output this help\n");
	printf("          -i <DEVICE>     select the device interface from which nSniff\n");
	printf("                          should sniff\n");
	printf("          -m <BYTES>      maximum packet length that should be captured\n");
	printf("                          at once (default: %d)\n", SNAPLEN);
	printf("          -n              prints names instead of numeric ports and addresses\n");
	printf("          -r <BYTES>      prints the whole packets as RAW data before parsing\n");
	printf("                          the headers\n");
	printf("                          headers\n");
	printf("          -v              run nSniff in verbose mode\n\n");
	 
	exit(1);
}

/* not much to comment */
void banner() {
	printf("\n[n]etw0rk [Sniff]er v0.1.2\n");
	printf("by Alexander Hauser (lex@lufix.net)\n\n");
	 
	return;
}

